__title__ = 'autossl'
__description__ = 'SSL certificates monitoring, renewal and deployment.'
__url__ = 'https://autossl.readthedocs.io'
__version__ = '0.9.0'
__author__ = 'Thibaud Castaing'
__author_email__ = 't-cas@users.noreply.github.com'
__license__ = 'MIT license'
__copyright__ = 'Copyright 2019 Amadeus S.A.S.'
